@@ -10,6 +10,10 @@
 #ifndef MSPACK_SYSTEM_H
 #define MSPACK_SYSTEM_H 1
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 /* ensure config.h is read before mspack.h */
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -53,16 +57,26 @@
 # include <limits.h>
 #endif
 
+#ifdef HAVE_INTTYPES_H
+# include <inttypes.h>
+#else
+# define PRId64 "lld"
+# define PRIu64 "llu"
+# define PRId32 "ld"
+# define PRIu32 "lu"
+#endif
+
 #if ((defined(_FILE_OFFSET_BITS) && _FILE_OFFSET_BITS >= 64) || \
      (defined(FILESIZEBITS)      && FILESIZEBITS      >= 64) || \
+     (defined(SIZEOF_OFF_T)      && SIZEOF_OFF_T      >= 8)  || \
      defined(_LARGEFILE_SOURCE) || defined(_LARGEFILE64_SOURCE))
 # define LARGEFILE_SUPPORT
-# define LD "lld"
-# define LU "llu"
+# define LD PRId64
+# define LU PRIu64
 #else
 extern const char *largefile_msg;
-# define LD "ld"
-# define LU "lu"
+# define LD PRId32
+# define LU PRIu32
 #endif
 
 /* endian-neutral reading of little-endian data */
@@ -112,4 +126,8 @@ static inline int mspack_memcmp(const void *s1, const void *s2, size_t n) {
 }
 #endif
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif
